@@ -1,6 +1,7 @@
 {-# LANGUAGE CPP        #-}
 {-# LANGUAGE MagicHash  #-}
 {-# LANGUAGE Rank2Types #-}
+{-# LANGUAGE BangPatterns #-}
 
 -----------------------------------------------------------------------------
 -- |
@@ -29,8 +30,11 @@ module Data.Serialize.Get (
     , runGetLazy
     , runGetState
     , runGetLazyState
+
+    -- ** Incremental interface
     , Result(..)
     , runGetPartial
+    , runGetChunk
 
     -- * Parsing
     , ensure
@@ -50,20 +54,31 @@ module Data.Serialize.Get (
 
     -- * Parsing particular types
     , getWord8
+    , getInt8
 
     -- ** ByteStrings
     , getByteString
     , getLazyByteString
 
+#if MIN_VERSION_bytestring(0,10,4)
+    , getShortByteString
+#endif
+
     -- ** Big-endian reads
     , getWord16be
     , getWord32be
     , getWord64be
+    , getInt16be
+    , getInt32be
+    , getInt64be
 
     -- ** Little-endian reads
     , getWord16le
     , getWord32le
     , getWord64le
+    , getInt16le
+    , getInt32le
+    , getInt64le
 
     -- ** Host-endian, unaligned reads
     , getWordhost
@@ -83,16 +98,19 @@ module Data.Serialize.Get (
     , getIntSetOf
     , getMaybeOf
     , getEitherOf
-
+    , getNested
   ) where
 
-import Control.Applicative (Applicative(..),Alternative(..))
-import Control.Monad (unless,when,ap,MonadPlus(..),liftM2)
+import qualified Control.Applicative as A
+import qualified Control.Monad as M
+import Control.Monad (unless)
+import qualified Control.Monad.Fail as Fail
 import Data.Array.IArray (IArray,listArray)
 import Data.Ix (Ix)
 import Data.List (intercalate)
 import Data.Maybe (isNothing,fromMaybe)
 import Foreign
+import System.IO.Unsafe (unsafeDupablePerformIO)
 
 import qualified Data.ByteString          as B
 import qualified Data.ByteString.Internal as B
@@ -106,6 +124,11 @@ import qualified Data.Set                 as Set
 import qualified Data.Tree                as T
 
 
+#if MIN_VERSION_bytestring(0,10,4)
+import qualified Data.ByteString.Short as BS
+#endif
+
+
 #if defined(__GLASGOW_HASKELL__) && !defined(__HADDOCK__)
 import GHC.Base
 import GHC.Word
@@ -143,8 +166,17 @@ newtype Get a = Get
 type Input  = B.ByteString
 type Buffer = Maybe B.ByteString
 
+emptyBuffer :: Buffer
+emptyBuffer  = Just B.empty
+
+extendBuffer :: Buffer -> B.ByteString -> Buffer
+extendBuffer buf chunk =
+  do bs <- buf
+     return $! bs `B.append` chunk
+{-# INLINE extendBuffer #-}
+
 append :: Buffer -> Buffer -> Buffer
-append l r = B.append `fmap` l <*> r
+append l r = B.append `fmap` l A.<*> r
 {-# INLINE append #-}
 
 bufferBytes :: Buffer -> B.ByteString
@@ -166,37 +198,62 @@ moreLength m = case m of
   Incomplete mb -> fromMaybe 0 mb
 
 instance Functor Get where
-    fmap p m =
-      Get $ \s0 b0 m0 kf ks ->
-        let ks' s1 b1 m1 a = ks s1 b1 m1 (p a)
-         in unGet m s0 b0 m0 kf ks'
+    fmap p m =        Get $ \ s0 b0 m0 kf ks ->
+      unGet m s0 b0 m0 kf $ \ s1 b1 m1 a     -> ks s1 b1 m1 (p a)
 
-instance Applicative Get where
-    pure  = return
-    (<*>) = ap
+instance A.Applicative Get where
+    pure a = Get $ \ s0 b0 m0 _ ks -> ks s0 b0 m0 a
+    {-# INLINE pure #-}
 
-instance Alternative Get where
+    f <*> x =         Get $ \ s0 b0 m0 kf ks ->
+      unGet f s0 b0 m0 kf $ \ s1 b1 m1 g     ->
+      unGet x s1 b1 m1 kf $ \ s2 b2 m2 y     -> ks s2 b2 m2 (g y)
+    {-# INLINE (<*>) #-}
+
+    m *> k =          Get $ \ s0 b0 m0 kf ks ->
+      unGet m s0 b0 m0 kf $ \ s1 b1 m1 _     -> unGet k s1 b1 m1 kf ks
+    {-# INLINE (*>) #-}
+
+instance A.Alternative Get where
     empty = failDesc "empty"
-    (<|>) = mplus
+    {-# INLINE empty #-}
+
+    (<|>) = M.mplus
+    {-# INLINE (<|>) #-}
 
 -- Definition directly from Control.Monad.State.Strict
 instance Monad Get where
-    return a = Get $ \ s0 b0 m0 _ ks -> ks s0 b0 m0 a
+    return = A.pure
+    {-# INLINE return #-}
 
-    m >>= g  =
-      Get $ \s0 b0 m0 kf ks ->
-        let ks' s1 b1 m1 a = unGet (g a) s1 b1 m1 kf ks
-         in unGet m s0 b0 m0 kf ks'
+    m >>= g  =        Get $ \ s0 b0 m0 kf ks ->
+      unGet m s0 b0 m0 kf $ \ s1 b1 m1 a     -> unGet (g a) s1 b1 m1 kf ks
+    {-# INLINE (>>=) #-}
 
+    (>>) = (A.*>)
+    {-# INLINE (>>) #-}
+
+    fail     = Fail.fail
+    {-# INLINE fail #-}
+
+instance Fail.MonadFail Get where
     fail     = failDesc
+    {-# INLINE fail #-}
 
-instance MonadPlus Get where
+instance M.MonadPlus Get where
     mzero     = failDesc "mzero"
+    {-# INLINE mzero #-}
+
     mplus a b =
       Get $ \s0 b0 m0 kf ks ->
-        let kf' _ b1 m1 _ _ = unGet b (s0 `B.append` bufferBytes b1)
-                                      (b0 `append` b1) m1 kf ks
-         in unGet a s0 (Just B.empty) m0 kf' ks
+        let ks' s1 b1        = ks s1 (b0 `append` b1)
+            kf' _  b1 m1     = kf (s0 `B.append` bufferBytes b1)
+                                  (b0 `append` b1) m1
+            try _  b1 m1 _ _ = unGet b (s0 `B.append` bufferBytes b1)
+                                       b1 m1 kf' ks'
+         in unGet a s0 emptyBuffer m0 try ks'
+    {-# INLINE mplus #-}
+
 
 ------------------------------------------------------------------------
 
@@ -234,10 +291,17 @@ runGet m str =
     Partial{} -> Left "Failed reading: Internal error: unexpected Partial."
 {-# INLINE runGet #-}
 
+-- | Run the get monad on a single chunk, providing an optional length for the
+-- remaining, unseen input, with Nothing indicating that it's not clear how much
+-- input is left.  For example, with a lazy ByteString, the optional length
+-- represents the sum of the lengths of all remaining chunks.
+runGetChunk :: Get a -> Maybe Int -> B.ByteString -> Result a
+runGetChunk m mbLen str = unGet m str Nothing (Incomplete mbLen) failK finalK
+{-# INLINE runGetChunk #-}
+
 -- | Run the Get monad applies a 'get'-based parser on the input ByteString
 runGetPartial :: Get a -> B.ByteString -> Result a
-runGetPartial m str =
-  unGet m str Nothing (Incomplete Nothing) failK finalK
+runGetPartial m = runGetChunk m Nothing
 {-# INLINE runGetPartial #-}
 
 -- | Run the Get monad applies a 'get'-based parser on the input
@@ -267,8 +331,15 @@ runGetState' m str off =
 -- Lazy Get --------------------------------------------------------------------
 
 runGetLazy' :: Get a -> L.ByteString -> (Either String a,L.ByteString)
-runGetLazy' m lstr = loop (Partial (runGetPartial m)) (L.toChunks lstr)
+runGetLazy' m lstr =
+  case L.toChunks lstr of
+    [c]  -> wrapStrict (runGetState' m c       0)
+    []   -> wrapStrict (runGetState' m B.empty 0)
+    c:cs -> loop (runGetChunk m (Just (len - B.length c)) c) cs
   where
+  len = fromIntegral (L.length lstr)
+
+  wrapStrict (e,s) = (e,L.fromChunks [s])
 
   loop result chunks = case result of
 
@@ -299,26 +370,50 @@ runGetLazyState m lstr = case runGetLazy' m lstr of
 
 -- | If at least @n@ bytes of input are available, return the current
 --   input, otherwise fail.
-ensure :: Int -> Get B.ByteString
-ensure n = n `seq` Get $ \ s0 b0 m0 kf ks ->
-    if B.length s0 >= n
-    then ks s0 b0 m0 s0
-    else unGet (demandInput >> ensureRec n) s0 b0 m0 kf ks
 {-# INLINE ensure #-}
+ensure :: Int -> Get B.ByteString
+ensure n0 = n0 `seq` Get $ \ s0 b0 m0 kf ks -> let
+    n' = n0 - B.length s0
+    in if n' <= 0
+        then ks s0 b0 m0 s0
+        else getMore n' s0 [] b0 m0 kf ks
+    where
+        -- The "accumulate and concat" pattern here is important not to incur
+        -- in quadratic behavior, see <https://github.com/GaloisInc/cereal/issues/48>
 
--- | If at least @n@ bytes of input are available, return the current
---   input, otherwise fail.
-ensureRec :: Int -> Get B.ByteString
-ensureRec n = Get $ \s0 b0 m0 kf ks ->
-    if B.length s0 >= n
-    then ks s0 b0 m0 s0
-    else unGet (demandInput >> ensureRec n) s0 b0 m0 kf ks
+        finalInput s0 ss = B.concat (reverse (s0 : ss))
+        finalBuffer b0 s0 ss = extendBuffer b0 (B.concat (reverse (init (s0 : ss))))
+
+        getMore !n s0 ss b0 m0 kf ks = let
+            tooFewBytes = let
+                !s = finalInput s0 ss
+                !b = finalBuffer b0 s0 ss
+                in kf s b m0 ["demandInput"] "too few bytes"
+            in case m0 of
+                Complete -> tooFewBytes
+                Incomplete mb -> Partial $ \s ->
+                    if B.null s
+                        then tooFewBytes
+                        else let
+                            !mb' = case mb of
+                                Just l -> Just $! l - B.length s
+                                Nothing -> Nothing
+                            in checkIfEnough n s (s0 : ss) b0 (Incomplete mb') kf ks
+
+        checkIfEnough !n s0 ss b0 m0 kf ks = let
+            n' = n - B.length s0
+            in if n' <= 0
+                then let
+                    !s = finalInput s0 ss
+                    !b = finalBuffer b0 s0 ss
+                    in ks s b m0 s
+                else getMore n' s0 ss b0 m0 kf ks
 
 -- | Isolate an action to operating within a fixed block of bytes.  The action
 --   is required to consume all the bytes that it is isolated to.
 isolate :: Int -> Get a -> Get a
 isolate n m = do
-  when (n < 0) (fail "Attempted to isolate a negative number of bytes")
+  M.when (n < 0) (fail "Attempted to isolate a negative number of bytes")
   s <- ensure n
   let (s',rest) = B.splitAt n s
   put s'
@@ -328,20 +423,6 @@ isolate n m = do
   put rest
   return a
 
--- | Immediately demand more input via a 'Partial' continuation
---   result.
-demandInput :: Get ()
-demandInput = Get $ \s0 b0 m0 kf ks ->
-  case m0 of
-    Complete      -> kf s0 b0 m0 ["demandInput"] "too few bytes"
-    Incomplete mb -> Partial $ \s ->
-      if B.null s
-      then kf s0 b0 m0 ["demandInput"] "too few bytes"
-      else let update l = l - B.length s
-               s1 = s0 `B.append` s
-               b1 = b0 `append` Just s
-            in ks s1 b1 (Incomplete (update `fmap` mb)) ()
-
 failDesc :: String -> Get a
 failDesc err = do
     let msg = "Failed reading: " ++ err
@@ -363,8 +444,11 @@ uncheckedSkip n = do
 -- Fails if @ga@ fails.
 lookAhead :: Get a -> Get a
 lookAhead ga = Get $ \ s0 b0 m0 kf ks ->
-  let ks' _s1 b1 = ks (s0 `B.append` bufferBytes b1) (b0 `append` b1)
-   in unGet ga s0 (Just B.empty) m0 kf ks'
+  -- the new continuation extends the old input with the new buffered bytes, and
+  -- appends the new buffer to the old one, if there was one.
+  let ks' _ b1 = ks (s0 `B.append` bufferBytes b1) (b0 `append` b1)
+      kf' _ b1 = kf s0 (b0 `append` b1)
+   in unGet ga s0 emptyBuffer m0 kf' ks'
 
 -- | Like 'lookAhead', but consume the input if @gma@ returns 'Just _'.
 -- Fails if @gma@ fails.
@@ -372,7 +456,7 @@ lookAheadM :: Get (Maybe a) -> Get (Maybe a)
 lookAheadM gma = do
     s <- get
     ma <- gma
-    when (isNothing ma) (put s)
+    M.when (isNothing ma) (put s)
     return ma
 
 -- | Like 'lookAhead', but consume the input if @gea@ returns 'Right _'.
@@ -425,6 +509,13 @@ getLazyByteString :: Int64 -> Get L.ByteString
 getLazyByteString n = f `fmap` getByteString (fromIntegral n)
   where f bs = L.fromChunks [bs]
 
+#if MIN_VERSION_bytestring(0,10,4)
+getShortByteString :: Int -> Get BS.ShortByteString
+getShortByteString n = do
+  bs <- getBytes n
+  return $! BS.toShort bs
+#endif
+
 
 ------------------------------------------------------------------------
 -- Helpers
@@ -453,72 +544,156 @@ getPtr :: Storable a => Int -> Get a
 getPtr n = do
     (fp,o,_) <- B.toForeignPtr `fmap` getBytes n
     let k p = peek (castPtr (p `plusPtr` o))
-    return (B.inlinePerformIO (withForeignPtr fp k))
+    return (unsafeDupablePerformIO (withForeignPtr fp k))
 {-# INLINE getPtr #-}
 
+-----------------------------------------------------------------------
+
+-- | Read a Int8 from the monad state
+getInt8 :: Get Int8
+getInt8 = do
+    s <- getBytes 1
+    return $! fromIntegral (B.unsafeHead s)
+
+-- | Read a Int16 in big endian format
+getInt16be :: Get Int16
+getInt16be = do
+    s <- getBytes 2
+    return $! (fromIntegral (s `B.unsafeIndex` 0) `shiftL` 8) .|.
+              (fromIntegral (s `B.unsafeIndex` 1) )
+
+-- | Read a Int16 in little endian format
+getInt16le :: Get Int16
+getInt16le = do
+    s <- getBytes 2
+    return $! (fromIntegral (s `B.unsafeIndex` 1) `shiftL` 8) .|.
+              (fromIntegral (s `B.unsafeIndex` 0) )
+
+-- | Read a Int32 in big endian format
+getInt32be :: Get Int32
+getInt32be = do
+    s <- getBytes 4
+    return $! (fromIntegral (s `B.unsafeIndex` 0) `shiftL` 24) .|.
+              (fromIntegral (s `B.unsafeIndex` 1) `shiftL` 16) .|.
+              (fromIntegral (s `B.unsafeIndex` 2) `shiftL`  8) .|.
+              (fromIntegral (s `B.unsafeIndex` 3) )
+
+-- | Read a Int32 in little endian format
+getInt32le :: Get Int32
+getInt32le = do
+    s <- getBytes 4
+    return $! (fromIntegral (s `B.unsafeIndex` 3) `shiftL` 24) .|.
+              (fromIntegral (s `B.unsafeIndex` 2) `shiftL` 16) .|.
+              (fromIntegral (s `B.unsafeIndex` 1) `shiftL`  8) .|.
+              (fromIntegral (s `B.unsafeIndex` 0) )
+
+-- | Read a Int64 in big endian format
+getInt64be :: Get Int64
+getInt64be = do
+    s <- getBytes 8
+    return $! (fromIntegral (s `B.unsafeIndex` 0) `shiftL` 56) .|.
+              (fromIntegral (s `B.unsafeIndex` 1) `shiftL` 48) .|.
+              (fromIntegral (s `B.unsafeIndex` 2) `shiftL` 40) .|.
+              (fromIntegral (s `B.unsafeIndex` 3) `shiftL` 32) .|.
+              (fromIntegral (s `B.unsafeIndex` 4) `shiftL` 24) .|.
+              (fromIntegral (s `B.unsafeIndex` 5) `shiftL` 16) .|.
+              (fromIntegral (s `B.unsafeIndex` 6) `shiftL`  8) .|.
+              (fromIntegral (s `B.unsafeIndex` 7) )
+
+-- | Read a Int64 in little endian format
+getInt64le :: Get Int64
+getInt64le = do
+    s <- getBytes 8
+    return $! (fromIntegral (s `B.unsafeIndex` 7) `shiftL` 56) .|.
+              (fromIntegral (s `B.unsafeIndex` 6) `shiftL` 48) .|.
+              (fromIntegral (s `B.unsafeIndex` 5) `shiftL` 40) .|.
+              (fromIntegral (s `B.unsafeIndex` 4) `shiftL` 32) .|.
+              (fromIntegral (s `B.unsafeIndex` 3) `shiftL` 24) .|.
+              (fromIntegral (s `B.unsafeIndex` 2) `shiftL` 16) .|.
+              (fromIntegral (s `B.unsafeIndex` 1) `shiftL`  8) .|.
+              (fromIntegral (s `B.unsafeIndex` 0) )
+
+{-# INLINE getInt8    #-}
+{-# INLINE getInt16be #-}
+{-# INLINE getInt16le #-}
+{-# INLINE getInt32be #-}
+{-# INLINE getInt32le #-}
+{-# INLINE getInt64be #-}
+{-# INLINE getInt64le #-}
+
 ------------------------------------------------------------------------
 
 -- | Read a Word8 from the monad state
 getWord8 :: Get Word8
-getWord8 = getPtr (sizeOf (undefined :: Word8))
+getWord8 = do
+    s <- getBytes 1
+    return (B.unsafeHead s)
 
 -- | Read a Word16 in big endian format
 getWord16be :: Get Word16
 getWord16be = do
     s <- getBytes 2
-    return $! (fromIntegral (s `B.index` 0) `shiftl_w16` 8) .|.
-              (fromIntegral (s `B.index` 1))
+    return $! (fromIntegral (s `B.unsafeIndex` 0) `shiftl_w16` 8) .|.
+              (fromIntegral (s `B.unsafeIndex` 1))
 
 -- | Read a Word16 in little endian format
 getWord16le :: Get Word16
 getWord16le = do
     s <- getBytes 2
-    return $! (fromIntegral (s `B.index` 1) `shiftl_w16` 8) .|.
-              (fromIntegral (s `B.index` 0) )
+    return $! (fromIntegral (s `B.unsafeIndex` 1) `shiftl_w16` 8) .|.
+              (fromIntegral (s `B.unsafeIndex` 0) )
 
 -- | Read a Word32 in big endian format
 getWord32be :: Get Word32
 getWord32be = do
     s <- getBytes 4
-    return $! (fromIntegral (s `B.index` 0) `shiftl_w32` 24) .|.
-              (fromIntegral (s `B.index` 1) `shiftl_w32` 16) .|.
-              (fromIntegral (s `B.index` 2) `shiftl_w32`  8) .|.
-              (fromIntegral (s `B.index` 3) )
+    return $! (fromIntegral (s `B.unsafeIndex` 0) `shiftl_w32` 24) .|.
+              (fromIntegral (s `B.unsafeIndex` 1) `shiftl_w32` 16) .|.
+              (fromIntegral (s `B.unsafeIndex` 2) `shiftl_w32`  8) .|.
+              (fromIntegral (s `B.unsafeIndex` 3) )
 
 -- | Read a Word32 in little endian format
 getWord32le :: Get Word32
 getWord32le = do
     s <- getBytes 4
-    return $! (fromIntegral (s `B.index` 3) `shiftl_w32` 24) .|.
-              (fromIntegral (s `B.index` 2) `shiftl_w32` 16) .|.
-              (fromIntegral (s `B.index` 1) `shiftl_w32`  8) .|.
-              (fromIntegral (s `B.index` 0) )
+    return $! (fromIntegral (s `B.unsafeIndex` 3) `shiftl_w32` 24) .|.
+              (fromIntegral (s `B.unsafeIndex` 2) `shiftl_w32` 16) .|.
+              (fromIntegral (s `B.unsafeIndex` 1) `shiftl_w32`  8) .|.
+              (fromIntegral (s `B.unsafeIndex` 0) )
 
 -- | Read a Word64 in big endian format
 getWord64be :: Get Word64
 getWord64be = do
     s <- getBytes 8
-    return $! (fromIntegral (s `B.index` 0) `shiftl_w64` 56) .|.
-              (fromIntegral (s `B.index` 1) `shiftl_w64` 48) .|.
-              (fromIntegral (s `B.index` 2) `shiftl_w64` 40) .|.
-              (fromIntegral (s `B.index` 3) `shiftl_w64` 32) .|.
-              (fromIntegral (s `B.index` 4) `shiftl_w64` 24) .|.
-              (fromIntegral (s `B.index` 5) `shiftl_w64` 16) .|.
-              (fromIntegral (s `B.index` 6) `shiftl_w64`  8) .|.
-              (fromIntegral (s `B.index` 7) )
+    return $! (fromIntegral (s `B.unsafeIndex` 0) `shiftl_w64` 56) .|.
+              (fromIntegral (s `B.unsafeIndex` 1) `shiftl_w64` 48) .|.
+              (fromIntegral (s `B.unsafeIndex` 2) `shiftl_w64` 40) .|.
+              (fromIntegral (s `B.unsafeIndex` 3) `shiftl_w64` 32) .|.
+              (fromIntegral (s `B.unsafeIndex` 4) `shiftl_w64` 24) .|.
+              (fromIntegral (s `B.unsafeIndex` 5) `shiftl_w64` 16) .|.
+              (fromIntegral (s `B.unsafeIndex` 6) `shiftl_w64`  8) .|.
+              (fromIntegral (s `B.unsafeIndex` 7) )
 
 -- | Read a Word64 in little endian format
 getWord64le :: Get Word64
 getWord64le = do
     s <- getBytes 8
-    return $! (fromIntegral (s `B.index` 7) `shiftl_w64` 56) .|.
-              (fromIntegral (s `B.index` 6) `shiftl_w64` 48) .|.
-              (fromIntegral (s `B.index` 5) `shiftl_w64` 40) .|.
-              (fromIntegral (s `B.index` 4) `shiftl_w64` 32) .|.
-              (fromIntegral (s `B.index` 3) `shiftl_w64` 24) .|.
-              (fromIntegral (s `B.index` 2) `shiftl_w64` 16) .|.
-              (fromIntegral (s `B.index` 1) `shiftl_w64`  8) .|.
-              (fromIntegral (s `B.index` 0) )
+    return $! (fromIntegral (s `B.unsafeIndex` 7) `shiftl_w64` 56) .|.
+              (fromIntegral (s `B.unsafeIndex` 6) `shiftl_w64` 48) .|.
+              (fromIntegral (s `B.unsafeIndex` 5) `shiftl_w64` 40) .|.
+              (fromIntegral (s `B.unsafeIndex` 4) `shiftl_w64` 32) .|.
+              (fromIntegral (s `B.unsafeIndex` 3) `shiftl_w64` 24) .|.
+              (fromIntegral (s `B.unsafeIndex` 2) `shiftl_w64` 16) .|.
+              (fromIntegral (s `B.unsafeIndex` 1) `shiftl_w64`  8) .|.
+              (fromIntegral (s `B.unsafeIndex` 0) )
+
+{-# INLINE getWord8    #-}
+{-# INLINE getWord16be #-}
+{-# INLINE getWord16le #-}
+{-# INLINE getWord32be #-}
+{-# INLINE getWord32le #-}
+{-# INLINE getWord64be #-}
+{-# INLINE getWord64le #-}
 
 ------------------------------------------------------------------------
 -- Host-endian reads
@@ -575,7 +750,7 @@ shiftl_w64 = shiftL
 -- Containers ------------------------------------------------------------------
 
 getTwoOf :: Get a -> Get b -> Get (a,b)
-getTwoOf ma mb = liftM2 (,) ma mb
+getTwoOf ma mb = M.liftM2 (,) ma mb
 
 -- | Get a list in the following format:
 --   Word64 (big endian format)
@@ -585,7 +760,7 @@ getTwoOf ma mb = liftM2 (,) ma mb
 getListOf :: Get a -> Get [a]
 getListOf m = go [] =<< getWord64be
   where
-  go as 0 = return (reverse as)
+  go as 0 = return $! reverse as
   go as i = do x <- m
                x `seq` go (x:as) (i - 1)
 
@@ -597,7 +772,7 @@ getListOf m = go [] =<< getWord64be
 --   ...
 --   element n
 getIArrayOf :: (Ix i, IArray a e) => Get i -> Get e -> Get (a i e)
-getIArrayOf ix e = liftM2 listArray (getTwoOf ix ix) (getListOf e)
+getIArrayOf ix e = M.liftM2 listArray (getTwoOf ix ix) (getListOf e)
 
 -- | Get a sequence in the following format:
 --   Word64 (big endian format)
@@ -614,7 +789,7 @@ getSeqOf m = go Seq.empty =<< getWord64be
 
 -- | Read as a list of lists.
 getTreeOf :: Get a -> Get (T.Tree a)
-getTreeOf m = liftM2 T.Node m (getListOf (getTreeOf m))
+getTreeOf m = M.liftM2 T.Node m (getListOf (getTreeOf m))
 
 -- | Read as a list of pairs of key and element.
 getMapOf :: Ord k => Get k -> Get a -> Get (Map.Map k a)
@@ -651,3 +826,10 @@ getEitherOf ma mb = do
   case tag of
     0 -> Left  `fmap` ma
     _ -> Right `fmap` mb
+
+-- | Read in a length and then read a nested structure
+--   of that length.
+getNested :: Get Int -> Get a -> Get a
+getNested getLen getVal = do
+    n <- getLen
+    isolate n getVal
